#include "curses.h"
#include <stdlib.h>
#include <signal.h>
#include "sorting.h"
#include <unistd.h>
#include <string.h>

extern void grab_info(bool filer_idle, const char *filter_userid);
extern void paint_top(WINDOW *w);
extern unsigned cpus;
extern void paint_cpu(WINDOW *w, bool total_only);
extern void paint_memory(WINDOW *w);
static void paint_cmd(WINDOW *w) {
	wclear(w);
}
extern void paint_header(WINDOW *w);
extern void paint_processes(WINDOW *w, bool filer_idle, const char *filter_user, Sorting_t sorting);

extern void do_kill(pid_t pid, int sig, WINDOW *err_window);
extern void do_renice(pid_t pid, int value, WINDOW *err_window);

static WINDOW *win_global=0;
static WINDOW *win_top=0;
static WINDOW *win_cpu=0;
static WINDOW *win_memory=0;
static WINDOW *win_cmd=0;
static WINDOW *win_header=0;
static WINDOW *win_processes=0;

static int refresh_interval=10;
static bool display_cpu_total_only = false;
static bool filter_idle=false;
static char filter_user[16+1]="";

static Sorting_t sorting = sorting_cpuusage;


static int getwinwidth(WINDOW *w) {
    int x1,x2;
    int y1,y2;
    getbegyx(w,y1,x1);
    getmaxyx(w,y2,x2);
    return x2-x1;
}
static int getwinheight(WINDOW *w) {
    int x1,x2;
    int y1,y2;
    getbegyx(w,y1,x1);
    getmaxyx(w,y2,x2);
    return y2-y1;
}

static void initialize_curses() {
	win_global = initscr();
	int win_global_width = getwinwidth(win_global);
	int win_global_height = getwinheight(win_global);
	int y=0;
	win_top = newwin(2,0,y,0);
	y+=2;
	if(display_cpu_total_only) {
		win_cpu = newwin(1,0,y,0);
		y++;
	} else {
		win_cpu = newwin(cpus,0,y,0);
		y+=cpus;
	}
	win_memory = newwin(2,0,y,0);
	y+=2;
	win_cmd = newwin(1,0,y,0);
	y+=1;
	win_header = newwin(1,0,y,0);
	y+=1;
//	win_processes = newwin(win_global_height-y,win_global_width,y,0);
printf("win_global_height=%d win_global_width=%d",win_global_height,win_global_width); fflush(stdout);
	win_processes = newwin(win_global_height-1,win_global_width,y,0);
}

static void finalize_curses() {
	delwin(win_top);
	delwin(win_cpu);
	delwin(win_memory);
	delwin(win_cmd);
	delwin(win_header);
	delwin(win_processes);
	endwin();
}

static void reinitialize_curses() {
	finalize_curses();
	
	initialize_curses();
	
	paint_top(win_top);
	paint_cpu(win_cpu,display_cpu_total_only);
	paint_memory(win_memory);
	paint_cmd(win_cmd);
	paint_header(win_header);
	paint_processes(win_processes,filter_idle,filter_user,sorting);
	wrefresh(win_global);
	wrefresh(win_top);
	wrefresh(win_cpu);
	wrefresh(win_memory);
	wrefresh(win_header);
	wrefresh(win_processes);
	wrefresh(win_cmd);
	refresh();
}

static bool sigwinch_received=false;
#ifdef __cplusplus
extern "C" {
#endif
static void sighandler_winch(int) {
	sigwinch_received=true;
    signal(SIGWINCH,sighandler_winch);
}
#ifdef __cplusplus
}
#endif

static bool prompt(const char *p, char *s, size_t sz) {
	wclear(win_cmd);
	wattron(win_cmd,A_BOLD);
	wprintw(win_cmd,(char*)"%s:",p);
	wattroff(win_cmd,A_BOLD);
	//wrefresh(win_cmd);
	
	int rc = wgetnstr(win_cmd,s,(int)sz);
	if(rc!=ERR) wclear(win_cmd);
	if(rc==ERR) return false;
	return true;
}

static void ui_refreshinterval() {
	char response[16];
	if(!prompt("Refresh interval",response,sizeof(response))) return;
	
	long l = strtol(response,0,10);
	if(l>0 && l<600)
		refresh_interval = (int)l;
}

static void ui_filteruser() {
	char response[16];
	if(!prompt("User (non=all)",response,sizeof(response))) return;
	
	strcpy(filter_user,response);
}

static bool prompt_pid_value(const char *pid_prompt, const char *value_prompt, pid_t *pid, int *value) {
	char pid_response[16];
	if(!prompt(pid_prompt,pid_response,sizeof(pid_response))) return false;
	if(pid_response[0]=='\0') return false;
	
	char *errpos=0;
	long l = strtol(pid_response,&errpos,10);
	if(errpos && *errpos) return false;
	*pid = (pid_t)l;
	
	char value_response[5];
	if(!prompt(value_prompt,value_response,sizeof(value_response))) return false;
	errpos=0;
	l = strtol(value_response,&errpos,10);
	if(errpos && *errpos) return false;
	if(value_response[0])
		*value = (int)l;
	return true;
}
	
static void ui_renice() {
	pid_t pid;
	static int value=0;
	char prompt_str[80];
	sprintf(prompt_str,"Change (default=%d)",value);
	if(!prompt_pid_value("Renice which process",prompt_str,&pid,&value)) return;
	do_renice(pid,value,win_cmd);
}

static void ui_kill() {
	pid_t pid;
	int value=SIGTERM;
	if(!prompt_pid_value("Kill which process","Change (default=SIGTERM)",&pid,&value)) return;
	do_kill(pid,value,win_cmd);
}


static void ui_help() {
	WINDOW *win_help = newwin(getwinheight(win_global),getwinwidth(win_global),0,0);
	wclear(win_help);
	wprintw(win_help,(char*)"Help for isj's hp-ux top\n");
	wprintw(win_help,(char*)"  h or ?   Display this page\n");
	wprintw(win_help,(char*)"  q        Quit\n");
	wprintw(win_help,(char*)"  i        Toggle display of idle processes\n");
	wprintw(win_help,(char*)"  c        Toggle display of per-cpu statistics\n");
	wprintw(win_help,(char*)"  s        Set refresh interval\n");
	wprintw(win_help,(char*)"  N        Sort by process id\n");
	wprintw(win_help,(char*)"  A        Sort by age\n");
	wprintw(win_help,(char*)"  P        Sort by CPU usage\n");
	wprintw(win_help,(char*)"  M        Sort by memory (resident)\n");
	wprintw(win_help,(char*)"  T        Sort by accumulative CPU time\n");
	wprintw(win_help,(char*)"  r        Renice a process\n");
	wprintw(win_help,(char*)"  k        Kill/signal a process\n");
	wprintw(win_help,(char*)"\n");
	wprintw(win_help,(char*)"Press any key to resume.");
	wrefresh(win_help);
	wgetch(win_help);
	delwin(win_help);
	refresh();
}

int main() {
printf("hej fra main\n");
	grab_info(filter_idle,filter_user);
	
	signal(SIGWINCH,sighandler_winch);
	initialize_curses();
	
	for(;;) {
		paint_top(win_top);
		paint_cpu(win_cpu,display_cpu_total_only);
		paint_memory(win_memory);
		paint_cmd(win_cmd);
        paint_header(win_header);
		paint_processes(win_processes,filter_idle,filter_user,sorting);

/*
		wrefresh(win_global);
		wrefresh(win_top);
		wrefresh(win_cpu);
		wrefresh(win_memory);
		//wrefresh(win_cmd);
		wrefresh(win_header);
		wrefresh(win_processes);
		
		wrefresh(win_cmd);
*/
		wnoutrefresh(win_top);
		wnoutrefresh(win_cpu);
		wnoutrefresh(win_memory);
    	wnoutrefresh(win_header);
		wnoutrefresh(win_processes);
		wnoutrefresh(win_cmd);
		doupdate();
		
		fd_set fds;
		FD_ZERO(&fds);
		FD_SET(0,&fds);
		struct timeval tv;
		tv.tv_sec = (long)refresh_interval;
		tv.tv_usec = 0;
		if(select(1,&fds,0,0,&tv)==1) {
			int c = wgetch(win_cmd);
			switch(c) {
				case 'q': goto finish;
				case 's':
                    ui_refreshinterval();
                    break;
				case 'i':
					filter_idle = !filter_idle;
					break;
				case 'u':
					ui_filteruser();
					break;
				case 'c':
					display_cpu_total_only = !display_cpu_total_only;
					initialize_curses();
					break;
				case 'N': sorting = sorting_pid; break;
				case 'A': sorting = sorting_age; break;
				case 'P': sorting = sorting_cpuusage; break;
				case 'M': sorting = sorting_resmem; break;
				case 'T': sorting = sorting_cputime; break;
				case 'r':
					ui_renice();
					break;
				case 'k':
					ui_kill();
					break;
				case '?':
				case 'h':
					ui_help();
					break;
                case ' ':
                    break; //refresh
                default:
                    continue;
			}
		}
		grab_info(filter_idle,filter_user);
        if(sigwinch_received) {
            sigwinch_received=false;
            reinitialize_curses();
        }
	}
finish:
	finalize_curses();
}
